#include "frmBitmap.h"

#include <Application.h>
#include <Region.h>
#include <OS.h>
#include <GraphicsDefs.h>
#include <Screen.h>

#include <new>

#include <memory.h>

#include "helper.h"

// view ///////////////////////////////////////////////////////////////////////////

FlipView::FlipView(FrameworkParams &params, status_t &error, BString &errorString) :

	BView(BRect(B_ORIGIN, params.frameBounds), NULL, B_FOLLOW_NONE, B_WILL_DRAW),
	
	mBack(NULL)
{
	if (B_OK != error)
		return;

	mMinimized = true;
	mNumRingBuf = params.numRingBuffers;
	mFlipSelector = 0;
	mDrawSelector = mNumRingBuf - 1;

	// check requested pixel format viability
	uint32 supportFlags = 0; 

	if (! ::bitmaps_support_space(params.pixelFormat, &supportFlags) ||
		! supportFlags & B_VIEWS_SUPPORT_DRAW_BITMAP)
	{
		errorString = "unsupported bitmap color space.";
		error = B_ERROR;
		return;
	}

	mBack = new(std::nothrow) BUFFER_RECORD [mNumRingBuf];
	if (NULL == mBack)
	{
		errorString = "'bad_alloc' at framework construction.";
		error = B_ERROR;
		return;
	}

	for (int32 i = 0; i < mNumRingBuf; i++)
	{
		mBack[i].writeLock = ::kreate_sem(0, NULL, error, errorString);
		mBack[i].unlockAtActivate = true;
		mBack[i].bitmap = new(std::nothrow) BBitmap(BRect(B_ORIGIN, params.frameBounds), params.pixelFormat);

		if (NULL == mBack[i].bitmap)
		{
			error = B_ERROR;
			errorString = "'bad_alloc' at framework construction.";
		}
	}

	if (B_OK != error) // account for last ::kreate_sem()
		return;

	SetViewColor(B_TRANSPARENT_COLOR);
}

FlipView::~FlipView()
{
	if (NULL != mBack)
	{
		for (int i = 0; i < mNumRingBuf; i++)
		{
			::delete_sem(mBack[i].writeLock);
			delete mBack[i].bitmap;
		}

		delete [] mBack;
	}
}

void FlipView::DestroyRingLocks()
{
	if (NULL != mBack)
		for (int i = 0; i < mNumRingBuf; i++)
		{
			::delete_sem(mBack[i].writeLock);
			mBack[i].writeLock = B_ERROR;
		}
}

void FlipView::MessageReceived(BMessage *message)
{
	switch (message->what)
	{
	case 'FLIP':

		mDrawSelector = mFlipSelector;

		if (!mMinimized)
		{
			BRegion clipper; 
			GetClippingRegion(&clipper);
		
			if (clipper.Frame().IsValid())
				for (int32 i = 0; i < clipper.CountRects(); i++)
					Draw(clipper.RectAt(i));

			::release_sem(mBack[mFlipSelector].writeLock);
		}
		else
			mBack[mFlipSelector].unlockAtActivate = true;

		++mFlipSelector %= mNumRingBuf;
		break;

	default:
		BView::MessageReceived(message);
	} 
}

void FlipView::Draw(BRect updateRect)
{
	DrawBitmap(mBack[mDrawSelector].bitmap, updateRect, updateRect);
}

void FlipView::WindowActivated(bool active)
{
	if (mMinimized != Window()->IsMinimized())
	{
		mMinimized = !mMinimized;

		if (!mMinimized)
			for (int i = 0; i < mNumRingBuf; i++)
				if (mBack[i].unlockAtActivate)
				{
					::release_sem_etc(mBack[i].writeLock, 1, B_DO_NOT_RESCHEDULE);
					mBack[i].unlockAtActivate = false;
				}
	}
}

// window /////////////////////////////////////////////////////////////////////

FrmBmpWindow::FrmBmpWindow(FrameworkParams &params, status_t &error, BString &errorString) :

	BWindow(BRect(B_ORIGIN, params.frameBounds), "BView_BBitmap Framework", B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE)
{
	mDrawSelector = 0;
	mView = NULL;

	if (B_OK != error)
		return;

	mView = new(std::nothrow) FlipView(params, error, errorString);
	if (NULL == mView)
	{
		errorString = "'bad_alloc' at framework construction.";
		error = B_ERROR;
		return;
	}

	AddChild(mView);

	BScreen *screen = new(std::nothrow) BScreen(this);
	if (NULL == screen)
	{
		errorString = "'bad_alloc' at framework construction.";
		error = B_ERROR;
		return;
	}

	BRect scrRect = screen->Frame();
	MoveTo((scrRect.Width() - params.frameBounds.x) * .5f, (scrRect.Height() - params.frameBounds.y) * .5f);
	delete screen;
}

FrmBmpWindow::~FrmBmpWindow()
{
	mView->DestroyRingLocks();

	int32 result;
	thread_id painter = ::find_thread(Framework::__painter_name());

	if (B_NAME_NOT_FOUND != painter)
		::wait_for_thread(painter, &result);

	// note:
	// all children of this window will be automatically deleted in BWindow's dtor.
}

bool FrmBmpWindow::QuitRequested()
{
	Hide(); 
	Sync();

 	::be_app_messenger.SendMessage(B_QUIT_REQUESTED);

	return BWindow::QuitRequested();
}

status_t FrmBmpWindow::frame_open(BUFFER_DESC &desc)
{
	status_t res = ::acquire_sem((*mView).mBack[mDrawSelector].writeLock);

	if (B_OK == res)
	{
		desc.frameHeight	= mView->mBack[mDrawSelector].bitmap->Bounds().IntegerHeight() + 1;
		desc.frameWidth		= mView->mBack[mDrawSelector].bitmap->Bounds().IntegerWidth() + 1;
		desc.pixelFormat	= mView->mBack[mDrawSelector].bitmap->ColorSpace();
		desc.rowStride		= mView->mBack[mDrawSelector].bitmap->BytesPerRow();
		desc.pBits			= mView->mBack[mDrawSelector].bitmap->Bits();
	}

	return res;
}

void FrmBmpWindow::frame_close()
{
	PostMessage('FLIP', mView);

	++mDrawSelector %= mView->mNumRingBuf;
}
