#include "frmNull.h"

#include <OS.h>
#include <Application.h>
#include <StringView.h>
#include <GraphicsDefs.h>
#include <Screen.h>

#include <new>
#include "helper.h"
#include "embossStringView.h"

#define LABEL_NODISPLAY_TEXT		"No display output in this framework."
#define	LABEL_NODISPLAY_FONTSIZE	17

// window /////////////////////////////////////////////////////////////////////

FrmNullWindow::FrmNullWindow(FrameworkParams &params, status_t &error, BString &errorString) : 

	BWindow(BRect(B_ORIGIN, params.frameBounds), "Null Framework", B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE),

	mDrawLock(B_ERROR),
	mBitsArea(B_ERROR),
	mBits(NULL)
{
	if (B_OK != error)
		return;

	mFrameWidth = (int32) params.frameBounds.x + 1;
	mFrameHeight = (int32) params.frameBounds.y + 1;
	mPixelFormat = params.pixelFormat;

	mDrawLock = ::kreate_sem(0, "drawing_lock", error, errorString);

	if (B_OK != error)
		return;

	// create a framebuffer of appropriate size
	size_t pixel_chunk;
	size_t pixels_per_chunk;
	size_t row_alignment;

	::get_pixel_size_for(mPixelFormat, &pixel_chunk, &row_alignment, &pixels_per_chunk);	

	mBytesPerRow = (mFrameWidth * pixel_chunk) / pixels_per_chunk;

	mBitsArea = ::kreate_area("frame_buffer", &mBits, B_ANY_ADDRESS,
							  (mFrameHeight * mBytesPerRow + B_PAGE_SIZE - 1) & ((uint32) -1 ^ (B_PAGE_SIZE - 1)),
							  B_NO_LOCK, B_READ_AREA | B_WRITE_AREA,
							  error, errorString);
	if (B_OK != error)
		return;

	// background view
	BView *background = new(std::nothrow) BView(BRect(B_ORIGIN, params.frameBounds), NULL,
												B_FOLLOW_LEFT | B_FOLLOW_TOP, B_WILL_DRAW);
	if (NULL == background)
	{
		errorString = "'bad_alloc' at framework construction.";
		error = B_ERROR;
		return;
	}

	(*background).SetViewColor(::ui_color(B_PANEL_BACKGROUND_COLOR));

	// print out a 'no actual output' message
	EmbossStringView *label = new(std::nothrow) EmbossStringView(BRect(0, 0, 0, 0), NULL, LABEL_NODISPLAY_TEXT);
	if (NULL == label)
	{
		errorString = "'bad_alloc' at framework construction.";
		error = B_ERROR;
		return;
	}

//	(*label).SetFontSize(LABEL_NODISPLAY_FONTSIZE);
	(*label).ResizeToPreferred();
	(*label).MoveTo((params.frameBounds.x - (*label).Bounds().Width()) * .5,
					(params.frameBounds.y - (*label).Bounds().Height()) * .5);

	(*label).SetHighColor(::tint_color(::ui_color(B_MENU_ITEM_TEXT_COLOR), B_DISABLED_MARK_TINT));
	(*label).SetLowColor(::ui_color(B_PANEL_BACKGROUND_COLOR));
	(*label).SetViewColor(::ui_color(B_PANEL_BACKGROUND_COLOR));
	(*label).SetAlignment(B_ALIGN_CENTER);

	(*background).AddChild(label);
	AddChild(background);

	// center the framework window on screen
	BScreen *screen = new(std::nothrow) BScreen(this);
	if (NULL == screen)
	{
		errorString = "'bad_alloc' at framework construction.";
		error = B_ERROR;
		return;
	}

	BRect scrRect = screen->Frame();
	MoveTo((scrRect.Width() - params.frameBounds.x) * .5f, (scrRect.Height() - params.frameBounds.y) * .5f);
	delete screen;
}

FrmNullWindow::~FrmNullWindow()
{
	// note: any children of this window will be automatically deleted on BWindow's dtor.

	::delete_sem(mDrawLock);
	
	int32 result; 
	thread_id painter = ::find_thread(Framework::__painter_name());

	if (B_NAME_NOT_FOUND != painter)
		::wait_for_thread(painter, &result);

	::delete_area(mBitsArea);
}

void FrmNullWindow::WindowActivated(bool active)
{
	if (active)
		::release_sem(mDrawLock);
	else
		::acquire_sem(mDrawLock);
}

bool FrmNullWindow::QuitRequested()
{
	Hide(); 
	Sync();

 	::be_app_messenger.SendMessage(B_QUIT_REQUESTED);

	return BWindow::QuitRequested();
}

status_t FrmNullWindow::frame_open(BUFFER_DESC &desc)
{
	status_t res = ::acquire_sem(mDrawLock);

	if (B_OK == res)
	{
		desc.frameHeight	= mFrameHeight;
		desc.frameWidth		= mFrameWidth;
		desc.pixelFormat	= mPixelFormat;
		desc.rowStride		= mBytesPerRow;
		desc.pBits			= mBits;
	}

	return res;
}

void FrmNullWindow::frame_close()
{
	::release_sem(mDrawLock);
}
