#ifndef __RENDCLIPRECT__
#define __RENDCLIPRECT__

#include "rendBaseTypes.h"

// IntersectCol6()		: finds the intersection between a line segment and a vertical line
//		- p0,			const Vector6 *		: line segment 1st vertex,	input
//		- p1,			const Vector6 *		: line segment 2nd vertex,	input
//		- x,			float				: column to intersect with,	input
//		- p_intersect,	Vector6 *			: point of intersection,	output
// returns
//		nil
// note
//		- line segment must be non-vertical
//		- input pointers are not checked for validity
inline void IntersectCol6(const Vector6 *p0,
						  const Vector6 *p1,
						  float x,
						  Vector6 *p_intersect)
{
	Vector6 d;
	d[0] = (*p1)[0] - (*p0)[0];
	d[1] = (*p1)[1] - (*p0)[1];
	d[2] = (*p1)[2] - (*p0)[2];
	d[3] = (*p1)[3] - (*p0)[3];
	d[4] = (*p1)[4] - (*p0)[4];
	d[5] = (*p1)[5] - (*p0)[5];

	const float ox = x - (*(const Vertex2 *)p0).x;

	(*p_intersect)[0] = x;
	(*p_intersect)[1] = ox * d[1] / d[0] + (*p0)[1];
	(*p_intersect)[2] = ox * d[2] / d[0] + (*p0)[2];
	(*p_intersect)[3] = ox * d[3] / d[0] + (*p0)[3];
	(*p_intersect)[4] = ox * d[4] / d[0] + (*p0)[4];
	(*p_intersect)[5] = ox * d[5] / d[0] + (*p0)[5];
}

// IntersectRow6()		: finds the intersection between a line segment and a horizontal line
//		- p0,			const Vector6 *		: line segment 1st vertex,	input
//		- p1,			const Vector6 *		: line segment 2nd vertex,	input
//		- y,			float				: row to intersect with,	input
//		- p_intersect,	Vector6 *			: point of intersection,	output
// returns
//		nil
// note
//		- line segment must be non-horizontal
//		- input pointers are not checked for validity
inline void IntersectRow6(const Vector6 *p0,
						  const Vector6 *p1,
						  float y,
						  Vector6 *p_intersect)
{
	Vector6 d;
	d[0] = (*p1)[0] - (*p0)[0];
	d[1] = (*p1)[1] - (*p0)[1];
	d[2] = (*p1)[2] - (*p0)[2];
	d[3] = (*p1)[3] - (*p0)[3];
	d[4] = (*p1)[4] - (*p0)[4];
	d[5] = (*p1)[5] - (*p0)[5];

	const float oy = y - (*(const Vertex2 *)p0).y;

	(*p_intersect)[0] = oy * d[0] / d[1] + (*p0)[0];
	(*p_intersect)[1] = y;
	(*p_intersect)[2] = oy * d[2] / d[1] + (*p0)[2];
	(*p_intersect)[3] = oy * d[3] / d[1] + (*p0)[3];
	(*p_intersect)[4] = oy * d[4] / d[1] + (*p0)[4];
	(*p_intersect)[5] = oy * d[5] / d[1] + (*p0)[5];
}


// ClipByRow6_ge()	: clips a polygon by a row, halfspace of greater/equal y is 'inside'
//		- border,	const float			: halfspaces border,						input
//		- poly,		const Vector6 **	: array containing a polygon definition,	input/output
//		- start,	uint32 &			: polygon starting vertex in the array,		input/output
//		- size,		uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,	Vector6 *			: storage buffer for the new vertices,		input/output
//		- nbuff,	uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- the polygon can be either clockwise or counter-clockwise;
//		- input pointers are not checked for validity
inline void ClipByRow6_ge(const float border,
						  const Vector6 *poly[],
						  uint32 &start,
						  uint32 &size,
						  Vector6 buffer[],
						  uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		if ((*(const Vertex2 *)poly[currVert]).y < border) // is current vertex outside?
		{
			if ((*(const Vertex2 *)poly[prevVert]).y > border) // is previous vertex strictly inside?
			{
				IntersectRow6(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if ((*(const Vertex2 *)poly[currVert]).y > border &&	// is current vertex strictly inside and
				(*(const Vertex2 *)poly[prevVert]).y < border)		// is previous vertex outside?
			{
				IntersectRow6(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}


// ClipByCol6_ge()		: clips a polygon by a column, halfspace of greater/equal x is 'inside'
//		- border,		const float			: halfspaces border,						input
//		- poly,			const Vector6 **	: array containing a polygon definition,	input/output
//		- start,		uint32 &			: polygon starting vertex in the array,		input/output
//		- size,			uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,		Vector6 *			: storage buffer for the new vertices,		input/output
//		- nbuff,		uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- the polygon can be either clockwise or counter-clockwise;
//		- input pointers are not checked for validity
inline void ClipByCol6_ge(const float border,
						  const Vector6 *poly[],
						  uint32 &start,
						  uint32 &size,
						  Vector6 buffer[],
						  uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		if ((*(const Vertex2 *)poly[currVert]).x < border) // is current vertex outside?
		{
			if ((*(const Vertex2 *)poly[prevVert]).x > border) // is previous vertex strictly inside?
			{
				IntersectCol6(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if ((*(const Vertex2 *)poly[currVert]).x > border &&	// is current vertex strictly inside and
				(*(const Vertex2 *)poly[prevVert]).x < border)		// is previous vertex outside?
			{
				IntersectCol6(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}


// ClipByRow6_le()		: clips a polygon by a row, halfspace of less/equal y is 'inside'
//		- border,		const float			: halfspaces border,						input
//		- poly,			const Vector6 **	: array containing a polygon definition,	input/output
//		- start,		uint32 &			: polygon starting vertex in the array,		input/output
//		- size,			uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,		Vector6 *			: storage buffer for the new vertices,		input/output
//		- nbuff,		uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- the polygon can be either clockwise or counter-clockwise;
//		- input pointers are not checked for validity
inline void ClipByRow6_le(const float border,
						  const Vector6 *poly[],
						  uint32 &start,
						  uint32 &size,
						  Vector6 buffer[],
						  uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		if ((*(const Vertex2 *)poly[currVert]).y > border) // is current vertex outside?
		{
			if ((*(const Vertex2 *)poly[prevVert]).y < border) // is previous vertex strictly inside?
			{
				IntersectRow6(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if ((*(const Vertex2 *)poly[currVert]).y < border &&	// is current vertex strictly inside and
				(*(const Vertex2 *)poly[prevVert]).y > border)		// is previous vertex outside?
			{
				IntersectRow6(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}


// ClipByCol6_le()	: clips a polygon by a column, halfspace of less/equal x is 'inside'
//		- border,	const float			: halfspaces border,						input
//		- poly,		const Vector6 **	: array containing a polygon definition,	input/output
//		- start,	uint32 &			: polygon starting vertex in the array,		input/output
//		- size,		uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,	Vector6 *			: storage buffer for the new vertices,		input/output
//		- nbuff,	uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- the polygon can be either clockwise or counter-clockwise;
//		- input pointers are not checked for validity
inline void ClipByCol6_le(const float border,
						  const Vector6 *poly[],
						  uint32 &start,
						  uint32 &size,
						  Vector6 buffer[],
						  uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		if ((*(const Vertex2 *)poly[currVert]).x > border) // is current vertex outside?
		{
			if ((*(const Vertex2 *)poly[prevVert]).x < border) // is previous vertex strictly inside?
			{
				IntersectCol6(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if ((*(const Vertex2 *)poly[currVert]).x < border &&	// is current vertex strictly inside and
				(*(const Vertex2 *)poly[prevVert]).x > border)		// is previous vertex outside?
			{
				IntersectCol6(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}


// ClipByRect6()	: clips a polygon by a screen-space rectangle
//		- rect,		const Rect &		: clipping rectangle,						input
//		- poly,		const Vector6 **	: array containing a polygon definition,	input/output
//		- start,	uint32 &			: polygon starting vertex in the array,		input/output
//		- size,		uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,	Vector6 *			: storage buffer for the new vertices,		input/output
//		- nbuff,	uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- size would be 0 if clipping discarded all vertices
inline void ClipByRect6(const Rect &rect,
						const Vector6 *poly[],
						uint32 &start,
						uint32 &size,
						Vector6 buffer[],
						uint32 &nbuff)
{
	ClipByRow6_ge(rect.top,	   poly, start, size, buffer, nbuff);
	ClipByCol6_ge(rect.left,   poly, start, size, buffer, nbuff);
	ClipByRow6_le(rect.bottom, poly, start, size, buffer, nbuff);
	ClipByCol6_le(rect.right,  poly, start, size, buffer, nbuff);
}

#endif // __RENDCLIPRECT__
