#include <math.h>
#include "rendSpan.h"

// Span4UV_any_nop_any()	: scan-converts a horizontal span into the void
//		- p0,				const Vertex4UV &	: span 1st vertex,	input
//		- p1,				const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,		int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_any_nop_any(const Vertex4UV &p0,
						 const Vertex4UV &p1,
						 int8 *linebuffer)
{ }

// Span4UV_nearest_cpy_argb8888()	: scan-converts a horizontal span, featuring nearest texture sampling, op. COPY
//		- p0,						const Vertex4UV &	: span 1st vertex,	input
//		- p1,						const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,				int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_cpy_argb8888(const Vertex4UV &p0,
								  const Vertex4UV &p1,
								  int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.0;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx = renderState.trigrad_x[0];
    const float dw_dx = renderState.trigrad_x[1];
    const float du_dx = renderState.trigrad_x[2];
    const float dv_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	   = p0.z + (err_x0 + (float) i) * dz_dx;
        const float rwi	   = p0.w + (err_x0 + (float) i) * dw_dx;
        const float ui_rwi = p0.u + (err_x0 + (float) i) * du_dx;
        const float vi_rwi = p0.v + (err_x0 + (float) i) * dv_dx;

		const float wi = 1.0 / rwi;

        const float ui = ui_rwi * wi;
        const float vi = vi_rwi * wi;

		int snap_ui = (int) floorf(ui + .5);
		int snap_vi = (int) floorf(vi + .5);

        snap_ui %= renderState.tex.dimx;
        snap_vi %= renderState.tex.dimy;

		PixelARGB8888 tex = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														snap_vi * renderState.tex.row_stride +
														snap_ui * sizeof(PixelARGB8888));

		*(PixelARGB8888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB8888)) = tex;
    }
}

// Span4UV_nearest_cpy_argb1555()	: scan-converts a horizontal span, featuring nearest texture sampling, op. COPY
//		- p0,						const Vertex4UV &	: span 1st vertex,	input
//		- p1,						const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,				int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_cpy_argb1555(const Vertex4UV &p0,
								  const Vertex4UV &p1,
								  int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.0;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx = renderState.trigrad_x[0];
    const float dw_dx = renderState.trigrad_x[1];
    const float du_dx = renderState.trigrad_x[2];
    const float dv_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	   = p0.z + (err_x0 + (float) i) * dz_dx;
        const float rwi	   = p0.w + (err_x0 + (float) i) * dw_dx;
        const float ui_rwi = p0.u + (err_x0 + (float) i) * du_dx;
        const float vi_rwi = p0.v + (err_x0 + (float) i) * dv_dx;

		const float wi = 1.0 / rwi;

        const float ui = ui_rwi * wi;
        const float vi = vi_rwi * wi;

		int snap_ui = (int) floorf(ui + .5);
		int snap_vi = (int) floorf(vi + .5);

        snap_ui %= renderState.tex.dimx;
        snap_vi %= renderState.tex.dimy;

		PixelARGB8888 tex = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														snap_vi * renderState.tex.row_stride +
														snap_ui * sizeof(PixelARGB8888));
		PixelARGB1555 formatted;

		formatted.a = (uint8) tex.a >> 7;
		formatted.r = (uint8) tex.r >> 3;
		formatted.g = (uint8) tex.g >> 3;
		formatted.b = (uint8) tex.b >> 3;

		*(PixelARGB1555 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB1555)) = formatted;
    }
}

// Span4UV_nearest_cpy_rgb888()	: scan-converts a horizontal span, featuring nearest texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_cpy_rgb888(const Vertex4UV &p0,
								const Vertex4UV &p1,
								int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.0;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx = renderState.trigrad_x[0];
    const float dw_dx = renderState.trigrad_x[1];
    const float du_dx = renderState.trigrad_x[2];
    const float dv_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	   = p0.z + (err_x0 + (float) i) * dz_dx;
        const float rwi	   = p0.w + (err_x0 + (float) i) * dw_dx;
        const float ui_rwi = p0.u + (err_x0 + (float) i) * du_dx;
        const float vi_rwi = p0.v + (err_x0 + (float) i) * dv_dx;

		const float wi = 1.0 / rwi;

        const float ui = ui_rwi * wi;
        const float vi = vi_rwi * wi;

		int snap_ui = (int) floorf(ui + .5);
		int snap_vi = (int) floorf(vi + .5);

        snap_ui %= renderState.tex.dimx;
        snap_vi %= renderState.tex.dimy;

		PixelARGB8888 tex = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														snap_vi * renderState.tex.row_stride +
														snap_ui * sizeof(PixelARGB8888));
		PixelRGB888 formatted;

		formatted.r = (uint8) tex.r;
		formatted.g = (uint8) tex.g;
		formatted.b = (uint8) tex.b;

		*(PixelRGB888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB888)) = formatted;
    }
}

// Span4UV_nearest_cpy_rgb565()	: scan-converts a horizontal span, featuring nearest texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_cpy_rgb565(const Vertex4UV &p0,
								const Vertex4UV &p1,
								int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.0;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx = renderState.trigrad_x[0];
    const float dw_dx = renderState.trigrad_x[1];
    const float du_dx = renderState.trigrad_x[2];
    const float dv_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	   = p0.z + (err_x0 + (float) i) * dz_dx;
        const float rwi	   = p0.w + (err_x0 + (float) i) * dw_dx;
        const float ui_rwi = p0.u + (err_x0 + (float) i) * du_dx;
        const float vi_rwi = p0.v + (err_x0 + (float) i) * dv_dx;

		const float wi = 1.0 / rwi;

        const float ui = ui_rwi * wi;
        const float vi = vi_rwi * wi;

		int snap_ui = (int) floorf(ui + .5);
		int snap_vi = (int) floorf(vi + .5);

        snap_ui %= renderState.tex.dimx;
        snap_vi %= renderState.tex.dimy;

		PixelARGB8888 tex = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														snap_vi * renderState.tex.row_stride +
														snap_ui * sizeof(PixelARGB8888));
		PixelRGB565 formatted;

		formatted.r = (uint8) tex.r >> 3;
		formatted.g = (uint8) tex.g >> 2;
		formatted.b = (uint8) tex.b >> 3;

		*(PixelRGB565 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB565)) = formatted;
    }
}

// Span4UV_linear_cpy_argb8888()	: scan-converts a horizontal span, featuring linear texture sampling, op. COPY
//		- p0,						const Vertex4UV &	: span 1st vertex,	input
//		- p1,						const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,				int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_linear_cpy_argb8888(const Vertex4UV &p0,
								 const Vertex4UV &p1,
								 int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.0;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx = renderState.trigrad_x[0];
    const float dw_dx = renderState.trigrad_x[1];
    const float du_dx = renderState.trigrad_x[2];
    const float dv_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	   = p0.z + (err_x0 + (float) i) * dz_dx;
        const float rwi	   = p0.w + (err_x0 + (float) i) * dw_dx;
        const float ui_rwi = p0.u + (err_x0 + (float) i) * du_dx;
        const float vi_rwi = p0.v + (err_x0 + (float) i) * dv_dx;

		const float wi = 1.0 / rwi;

        const float ui = ui_rwi * wi;
        const float vi = vi_rwi * wi;

		const float snap_ui = floorf(ui);
		const float snap_vi = floorf(vi);

        const float mass_u1 = ui - snap_ui;
        const float mass_v1 = vi - snap_vi;

		const float mass00 = (1.0 - mass_u1) * (1.0 - mass_v1);
		const float mass01 = mass_u1		 * (1.0 - mass_v1);
		const float mass11 = mass_u1		 * mass_v1;
		const float mass10 = (1.0 - mass_u1) * mass_v1;

        int ui0 = (int) snap_ui;
        int vi0 = (int) snap_vi;
        int ui1 = ui0 + 1;
        int vi1 = vi0 + 1;

        ui0 %= renderState.tex.dimx;
        vi0 %= renderState.tex.dimy;
        ui1 %= renderState.tex.dimx;
        vi1 %= renderState.tex.dimy;

		PixelARGB8888 tex00 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex01 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 tex10 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex11 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 filtered;

		filtered.a = (uint8) floorf(.5 +
			(float)tex00.a * mass00 + (float)tex01.a * mass01 + (float)tex10.a * mass10 + (float)tex11.a * mass11);
		filtered.r = (uint8) floorf(.5 +
			(float)tex00.r * mass00 + (float)tex01.r * mass01 + (float)tex10.r * mass10 + (float)tex11.r * mass11);
		filtered.g = (uint8) floorf(.5 +
			(float)tex00.g * mass00 + (float)tex01.g * mass01 + (float)tex10.g * mass10 + (float)tex11.g * mass11);
		filtered.b = (uint8) floorf(.5 +
			(float)tex00.b * mass00 + (float)tex01.b * mass01 + (float)tex10.b * mass10 + (float)tex11.b * mass11);

		*(PixelARGB8888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB8888)) = filtered;
    }
}

// Span4UV_linear_cpy_argb1555()	: scan-converts a horizontal span, featuring linear texture sampling, op. COPY
//		- p0,						const Vertex4UV &	: span 1st vertex,	input
//		- p1,						const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,				int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_linear_cpy_argb1555(const Vertex4UV &p0,
								 const Vertex4UV &p1,
								 int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.0;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx = renderState.trigrad_x[0];
    const float dw_dx = renderState.trigrad_x[1];
    const float du_dx = renderState.trigrad_x[2];
    const float dv_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	   = p0.z + (err_x0 + (float) i) * dz_dx;
        const float rwi	   = p0.w + (err_x0 + (float) i) * dw_dx;
        const float ui_rwi = p0.u + (err_x0 + (float) i) * du_dx;
        const float vi_rwi = p0.v + (err_x0 + (float) i) * dv_dx;

		const float wi = 1.0 / rwi;

        const float ui = ui_rwi * wi;
        const float vi = vi_rwi * wi;

		const float snap_ui = floorf(ui);
		const float snap_vi = floorf(vi);

        const float mass_u1 = ui - snap_ui;
        const float mass_v1 = vi - snap_vi;

		const float mass00 = (1.0 - mass_u1) * (1.0 - mass_v1);
		const float mass01 = mass_u1		 * (1.0 - mass_v1);
		const float mass11 = mass_u1		 * mass_v1;
		const float mass10 = (1.0 - mass_u1) * mass_v1;

        int ui0 = (int) snap_ui;
        int vi0 = (int) snap_vi;
        int ui1 = ui0 + 1;
        int vi1 = vi0 + 1;

        ui0 %= renderState.tex.dimx;
        vi0 %= renderState.tex.dimy;
        ui1 %= renderState.tex.dimx;
        vi1 %= renderState.tex.dimy;

		PixelARGB8888 tex00 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex01 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 tex10 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex11 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB1555 filtered;

		filtered.a = (uint8) floorf(.5 +
			(float)tex00.a * mass00 + (float)tex01.a * mass01 + (float)tex10.a * mass10 + (float)tex11.a * mass11) >> 7;
		filtered.r = (uint8) floorf(.5 +
			(float)tex00.r * mass00 + (float)tex01.r * mass01 + (float)tex10.r * mass10 + (float)tex11.r * mass11) >> 3;
		filtered.g = (uint8) floorf(.5 +
			(float)tex00.g * mass00 + (float)tex01.g * mass01 + (float)tex10.g * mass10 + (float)tex11.g * mass11) >> 3;
		filtered.b = (uint8) floorf(.5 +
			(float)tex00.b * mass00 + (float)tex01.b * mass01 + (float)tex10.b * mass10 + (float)tex11.b * mass11) >> 3;

		*(PixelARGB1555 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB1555)) = filtered;
    }
}

// Span4UV_linear_cpy_rgb888()	: scan-converts a horizontal span, featuring linear texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_linear_cpy_rgb888(const Vertex4UV &p0,
							   const Vertex4UV &p1,
							   int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.0;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx = renderState.trigrad_x[0];
    const float dw_dx = renderState.trigrad_x[1];
    const float du_dx = renderState.trigrad_x[2];
    const float dv_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	   = p0.z + (err_x0 + (float) i) * dz_dx;
        const float rwi	   = p0.w + (err_x0 + (float) i) * dw_dx;
        const float ui_rwi = p0.u + (err_x0 + (float) i) * du_dx;
        const float vi_rwi = p0.v + (err_x0 + (float) i) * dv_dx;

		const float wi = 1.0 / rwi;

        const float ui = ui_rwi * wi;
        const float vi = vi_rwi * wi;

		const float snap_ui = floorf(ui);
		const float snap_vi = floorf(vi);

        const float mass_u1 = ui - snap_ui;
        const float mass_v1 = vi - snap_vi;

		const float mass00 = (1.0 - mass_u1) * (1.0 - mass_v1);
		const float mass01 = mass_u1		 * (1.0 - mass_v1);
		const float mass11 = mass_u1		 * mass_v1;
		const float mass10 = (1.0 - mass_u1) * mass_v1;

        int ui0 = (int) snap_ui;
        int vi0 = (int) snap_vi;
        int ui1 = ui0 + 1;
        int vi1 = vi0 + 1;

        ui0 %= renderState.tex.dimx;
        vi0 %= renderState.tex.dimy;
        ui1 %= renderState.tex.dimx;
        vi1 %= renderState.tex.dimy;

		PixelARGB8888 tex00 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex01 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 tex10 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex11 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelRGB888 filtered;

		filtered.r = (uint8) floorf(.5 +
			(float)tex00.r * mass00 + (float)tex01.r * mass01 + (float)tex10.r * mass10 + (float)tex11.r * mass11);
		filtered.g = (uint8) floorf(.5 +
			(float)tex00.g * mass00 + (float)tex01.g * mass01 + (float)tex10.g * mass10 + (float)tex11.g * mass11);
		filtered.b = (uint8) floorf(.5 +
			(float)tex00.b * mass00 + (float)tex01.b * mass01 + (float)tex10.b * mass10 + (float)tex11.b * mass11);

		*(PixelRGB888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB888)) = filtered;
    }
}

// Span4UV_linear_cpy_rgb565()	: scan-converts a horizontal span, featuring linear texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_linear_cpy_rgb565(const Vertex4UV &p0,
							   const Vertex4UV &p1,
							   int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.0;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx = renderState.trigrad_x[0];
    const float dw_dx = renderState.trigrad_x[1];
    const float du_dx = renderState.trigrad_x[2];
    const float dv_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	   = p0.z + (err_x0 + (float) i) * dz_dx;
        const float rwi	   = p0.w + (err_x0 + (float) i) * dw_dx;
        const float ui_rwi = p0.u + (err_x0 + (float) i) * du_dx;
        const float vi_rwi = p0.v + (err_x0 + (float) i) * dv_dx;

		const float wi = 1.0 / rwi;

        const float ui = ui_rwi * wi;
        const float vi = vi_rwi * wi;

		const float snap_ui = floorf(ui);
		const float snap_vi = floorf(vi);

        const float mass_u1 = ui - snap_ui;
        const float mass_v1 = vi - snap_vi;

		const float mass00 = (1.0 - mass_u1) * (1.0 - mass_v1);
		const float mass01 = mass_u1		 * (1.0 - mass_v1);
		const float mass11 = mass_u1		 * mass_v1;
		const float mass10 = (1.0 - mass_u1) * mass_v1;

        int ui0 = (int) snap_ui;
        int vi0 = (int) snap_vi;
        int ui1 = ui0 + 1;
        int vi1 = vi0 + 1;

        ui0 %= renderState.tex.dimx;
        vi0 %= renderState.tex.dimy;
        ui1 %= renderState.tex.dimx;
        vi1 %= renderState.tex.dimy;

		PixelARGB8888 tex00 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex01 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 tex10 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex11 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelRGB565 filtered;

		filtered.r = (uint8) floorf(.5 +
			(float)tex00.r * mass00 + (float)tex01.r * mass01 + (float)tex10.r * mass10 + (float)tex11.r * mass11) >> 3;
		filtered.g = (uint8) floorf(.5 +
			(float)tex00.g * mass00 + (float)tex01.g * mass01 + (float)tex10.g * mass10 + (float)tex11.g * mass11) >> 2;
		filtered.b = (uint8) floorf(.5 +
			(float)tex00.b * mass00 + (float)tex01.b * mass01 + (float)tex10.b * mass10 + (float)tex11.b * mass11) >> 3;

		*(PixelRGB565 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB565)) = filtered;
    }
}
