#include <math.h>
#include "rendVect.h"

void Rotate(Matrix4 &mat, float a, float x, float y, float z)
{
	Vector4 v = { x, y, z, 0 };
	MulVect4(v, 1.0 / LenVect4(v));

	const float sin_a = sinf(a);
	const float cos_a = cosf(a);

	x = v[0];
	y = v[1];
	z = v[2];
	
	LoadMat(mat,
			x * x + cos_a * (1 - x * x),		 y * x - cos_a * (y * x) - sin_a * z, z * x - cos_a * (z * x) + sin_a * y, 0,
			x * y - cos_a * (x * y) + sin_a * z, y * y + cos_a * (1 - y * y),		  z * y - cos_a * (z * y) - sin_a * x, 0,
			x * z - cos_a * (x * z) - sin_a * y, y * z - cos_a * (y * z) + sin_a * x, z * z + cos_a * (1 - z * z),		   0,
			0,									 0,									  0,								   1);
}

void Translate(Matrix4 &mat, float x, float y, float z)
{
	LoadMat(mat,
			1,	0,	0,	x,
			0,	1,	0,	y,
			0,	0,	1,	z,
			0,	0,	0,	1);
}

void Scale(Matrix4 &mat, float x, float y, float z)
{
	LoadMat(mat,
			x,	0,	0,	0,
			0,	y,	0,	0,
			0,	0,	z,	0,
			0,	0,	0,	1);
}

// Frustum()	: builds a perspective-projection transform
//		- mat,	Matrix4 &	: matrix of the projection transform,								output
//		- l,	float		: left edge of viewport's mapping onto the near clipping plane,		input
//		- r,	float		: right edge of viewport's mapping onto the near clipping plane,	input
//		- b,	float		: bottom edge of viewport's mapping onto the near clipping plane,	input
//		- t,	float		: top edge of viewport's mapping onto the near clipping plane,		input
//		- n,	float		: absolute distance from the origin to the near clipping plane,		input
//		- f,	float		: absolute distance from the origin to the far clipping plane,		input
// returns
//		nil
// note
//		- camera space taken as right-handed, positive y-axis pointing up
//		  and negative z-axis pointing in the direction of the view vector
//		- the distances to the near and far clipping planes are in absolute values,
//		  whereas the two planes are at -n and -f units from the origin, respectively
//		- clipping is	-Wc <= Xc <= Wc,	and DNC is	-1 <= Xd <= 1
//						-Wc <= Yc <= Wc					-1 <= Yd <= 1
//						-Wc <= Zc <= Wc					-1 <= Zd <= 1
//		- projection plane equals the near clipping plane
void Frustum(Matrix4 &mat, float l, float r, float b, float t, float n, float f)
{
	LoadMat(mat,
			2 * n / (r - l),	0,					(r + l) / (r - l),	0,
			0,					2 * n / (t - b),	(t + b) / (t - b),	0,
			0,					0,				  - (f + n) / (f - n),	- 2 * f * n / (f - n),
			0,					0,				  - 1,					0);
}

// Ortho()		: builds a parallel-projection transform
//		- mat,	Matrix4 &	: matrix of the projection transform,								output
//		- l,	float		: left edge of viewport's mapping onto the near clipping plane,		input
//		- r,	float		: right edge of viewport's mapping onto the near clipping plane,	input
//		- b,	float		: bottom edge of viewport's mapping onto the near clipping plane,	input
//		- t,	float		: top edge of viewport's mapping onto the near clipping plane,		input
//		- n,	float		: absolute distance from the origin to the near clipping plane,		input
//		- f,	float		: absolute distance from the origin to the far clipping plane,		input
// returns
//		nil
// note
//		- camera space taken as right-handed, positive y-axis pointing up
//		  and negative z-axis pointing in the direction of the view vector
//		- the distances to the near and far clipping planes are in absolute values,
//		  whereas the two planes are at -n and -f units from the origin, respectively
//		- clipping is	-Wc <= Xc <= Wc,	and DNC is	-1 <= Xd <= 1
//						-Wc <= Yc <= Wc					-1 <= Yd <= 1
//						-Wc <= Zc <= Wc					-1 <= Zd <= 1
//		- projection plane equals the near clipping plane
void Ortho(Matrix4 &mat, float l, float r, float b, float t, float n, float f)
{
	LoadMat(mat,
			2 / (r - l),	0,				0,				- (r + l) / (r - l),
			0,				2 / (t - b), 	0,				- (t + b) / (t - b),
			0,				0,				- 2 / (f - n),	- (f + n) / (f - n),
			0,				0,				0,				1 );
}

// CopyInvertMat()	: finds the inverse of the given matrix
//		- mat,		const Matrix4 &	: argument matrix,								input
//		- inv,		Matrix4 &		: inverse of the argument,						output
// returns
//		nil
// note
//		- quick hack for orthogonal matrices only.
void CopyInvertMat(const Matrix4 &mat, Matrix4 &inv)
{
	LoadMat(inv,
			mat[0][0], mat[1][0], mat[2][0], - (mat[0][0] * mat[0][3] + mat[1][0] * mat[1][3] + mat[2][0] * mat[2][3]),
			mat[0][1], mat[1][1], mat[2][1], - (mat[0][1] * mat[0][3] + mat[1][1] * mat[1][3] + mat[2][1] * mat[2][3]),
			mat[0][2], mat[1][2], mat[2][2], - (mat[0][2] * mat[0][3] + mat[1][2] * mat[1][3] + mat[2][2] * mat[2][3]),
			0,		   0,		  0,		 1);
}
